from enum import IntEnum

class MessageLength(IntEnum):
    UNDETERMINED = 0
    COMMAND = 13
    RESPONSE = 395

class CommandMessageConstants(IntEnum):
    BYTE_0 = 85             # Byte 0 of command always fixed
    BYTE_2 = 12             # Byte 2 of command always fixed
    TOGGLE = 128            # toggle is common to both ACCommands and GroupCommands

# Go in byte 1 of command messages
class CommandMessageType(IntEnum):
    REQUEST_STATE = 1
    GROUP_CONTROL = 129
    AC_CONTROL = 134

# Go in byte 4 of CommandMessageType.AC_CONTROL messages
class ACCommands(IntEnum):
    SET_MODE = 129
    SET_FAN_SPEED = 130
    TEMP_DEC = 147
    TEMP_INC = 163

class GroupCommands(IntEnum):
    CHANGE_DAMP = 1      # 1 goes in byte 5 for change damp
    TOGGLE = 0           # 0 goes in byte 5 for toggle
    DAMP_DEC = 1         # Damper decrement (10% down) in byte 4
    DAMP_INC = 2         # Damper increment (10% up) in byte 4


class ResponseMessageConstants(IntEnum):
    LONG_STRING_LENGTH = 16
    SHORT_STRING_LENGTH = 8

# Thanks to home-assistant.io user Radebe2k
class ResponseMessageOffsets(IntEnum):
    # Header is 2 bytes
    HEADER = 0
    # There are 16 zones
    # zone names are 8 bytes (ResponseMessageConstants.SHORT_STRING_LENGTH)
    GROUP_NAMES_START = 100
    # zone statuses are 1 byte each
    ZONE_STATUSES_START = 228
    # In order, each group and the number of zones it consists of
    # Zones must be in consecutive order
    GROUP_ZONES_START = 244
    # Zone strengths are 1 byte each
    ZONE_DAMPS_START = 276
    NUM_GROUPS = 292
    TURBO_GROUP = 297
    # Contains isTurbo, isSafety, isSpill bits of the 2 ACs
    ACs_STATUS = 299
    TOUCHPAD_TEMP = 323 # I think?? Idk what happens when there's 2 touchpads then...
    # System name is 16 bytes (ResponseMessageConstants.LONG_STRING_LENGTH)
    SYSTEM_NAME = 324
    # On/Off, isError, 'Auto off' enabled, 'thermistor on AC', program number
    AC1_STATUS = 354
    #AC2_STATUS = 355
    AC1_BRAND = 356
    #AC2_BRAND = 357
    AC1_MODE = 358
    #AC2_MODE = 359
    AC1_FAN_SPEED = 360
    #AC2_FAN_SPEED = 361
    AC1_SET_TEMP = 362
    #AC2_SET_TEMP = 363
    AC1_MEASURED_TEMP = 364
    #AC2_MEASURED_TEMP = 365
    AC1_ERROR_CODE = 366
    #AC2_ERROR_CODE = 367
    AC1_GATEWAY_ID = 368
    #AC2_GATEWAY_ID = 369
    # AC names are 8 bytes (ResponseMessageConstants.SHORT_STRING_LENGTH)
    AC1_NAME_START = 370
    #AC2_NAME_START = 378
    HASH = 394



from airtouch2.protocol.messages.Message import Message
from airtouch2.protocol.constants import MessageLength, ResponseMessageConstants, ResponseMessageOffsets
from airtouch2.protocol.enums import ACBrand, ACMode

class ResponseMessage(Message):
    """ The airtouch 2 response message (there is only one) that contains all the information about the current state of the system"""
    length: MessageLength = MessageLength.RESPONSE

    # only getting things for AC1 for now
    def __init__(self, raw_response: bytes):
        super().__init__()

        status = raw_response[ResponseMessageOffsets.AC1_STATUS]
        # MS bit is on/off
        self.ac_active = [(status & 0x80 > 0)]
        # Bit 7 is whether AC is errored
        self.ac_error = [(status & 0x40 > 0)]
        self.ac_error_code = [raw_response[ResponseMessageOffsets.AC1_ERROR_CODE]]
        # bit 4 is whether or not 'thermistor on AC' is checked (0 = checked)
        self.ac_thermistor = [(status & 0x04 == 0)]
        # lowest 3 bits are AC program number
        #self.ac_program = [(status & 0x07)]

        status2 = raw_response[ResponseMessageOffsets.ACs_STATUS]
        self.ac_turbo = [(status2 & 0x20 > 0)]
        self.ac_safety = [(status2 & 0x04 > 0)]
        self.ac_spill = [(status2 & 0x02) > 0]

        # simply 0-4, see ACMode enum
        self.ac_mode = [ACMode(raw_response[ResponseMessageOffsets.AC1_MODE])]
        # most signficant byte is # of speeds, least significant byte is speed (the meaning of which depends), see AT2Aircon::_set_true_and_supported_fan_speed()
        self.ac_num_fan_speeds = [(raw_response[ResponseMessageOffsets.AC1_FAN_SPEED] & 0xF0) >> 4]
        self.ac_fan_speed = [raw_response[ResponseMessageOffsets.AC1_FAN_SPEED] & 0x0F]

        self.ac_set_temp = [raw_response[ResponseMessageOffsets.AC1_SET_TEMP]]
        self.ac_measured_temp = [raw_response[ResponseMessageOffsets.AC1_MEASURED_TEMP]]
        self.touchpad_temp = raw_response[ResponseMessageOffsets.TOUCHPAD_TEMP]

        self.ac_brand = [ACBrand(raw_response[ResponseMessageOffsets.AC1_BRAND])]
        self.ac_gateway_id = [raw_response[ResponseMessageOffsets.AC1_GATEWAY_ID]]

        self.ac_name = [raw_response[ResponseMessageOffsets.AC1_NAME_START:ResponseMessageOffsets.AC1_NAME_START+ResponseMessageConstants.SHORT_STRING_LENGTH].decode().split("\0")[0]]
        self.system_name = raw_response[ResponseMessageOffsets.SYSTEM_NAME:ResponseMessageOffsets.SYSTEM_NAME+ResponseMessageConstants.LONG_STRING_LENGTH].decode().split("\0")[0]

        # Group stuff
        self.num_groups = raw_response[ResponseMessageOffsets.NUM_GROUPS]
        # list of tuples: (start_zone, num_zones)
        self.group_zones = [((raw_response[offset] & 0xF0) >> 4, raw_response[offset] & 0x0F) for offset in range(ResponseMessageOffsets.GROUP_ZONES_START, ResponseMessageOffsets.GROUP_ZONES_START + 16)]
        self.zone_damps = [raw_response[offset] for offset in range(ResponseMessageOffsets.ZONE_DAMPS_START, ResponseMessageOffsets.ZONE_DAMPS_START+16)]
        self.group_names = [raw_response[offset:offset+ResponseMessageConstants.SHORT_STRING_LENGTH].decode().split("\0")[0] for offset in range(ResponseMessageOffsets.GROUP_NAMES_START, ResponseMessageOffsets.GROUP_NAMES_START+16*ResponseMessageConstants.SHORT_STRING_LENGTH, ResponseMessageConstants.SHORT_STRING_LENGTH)]
        zone_statuses = [raw_response[offset] for offset in range(ResponseMessageOffsets.ZONE_STATUSES_START, ResponseMessageOffsets.ZONE_STATUSES_START+16)]
        self.zone_ons = [(status & 0x80 > 0) for status in zone_statuses]
        self.zone_spills = [(status & 0x40 > 0) for status in zone_statuses]
        self.turbo_group = raw_response[ResponseMessageOffsets.TURBO_GROUP]



from __future__ import annotations
from itertools import compress
from typing import TYPE_CHECKING
from airtouch2.protocol.messages import ResponseMessage

from airtouch2.protocol.messages import ChangeDamper, ToggleGroup
if TYPE_CHECKING:
    from airtouch2.AT2Client import AT2Client
import logging


_LOGGER = logging.getLogger(__name__)

class AT2Group:
    def __init__(self, client: AT2Client, number: int, response: ResponseMessage):
        self._client = client
        self.number = number
        self.update(response)

    def update(self, response: ResponseMessage):
        self.name = response.group_names[self.number]
        [start_zone, num_zones] = response.group_zones[self.number]
        # 0 to 10 steps of 10%
        self.damp = response.zone_damps[start_zone]
        self.spill = response.zone_spills[start_zone]
        self.on = response.zone_ons[start_zone]
        mismatches: set[str] = set()
        for i in range(start_zone+1, start_zone + num_zones):
            # this group is spilling if any of its zones are
            self.spill = response.zone_spills[i]
            # these should match for all zones that comprise this group
            if (self.damp != response.zone_damps[i]):
                mismatches.add("open percents")
            if (self.on != response.zone_ons[i]):
                mismatches.add("on/offs")
            if mismatches:
                _LOGGER.warning(f"Zones of group '{self.name}' have mismatching {', '.join(mismatches)}")

        self.turbo = True if response.turbo_group == self.number else False

    def inc_dec_damp(self, inc: bool):
        self._client.send_command(ChangeDamper(self.number, inc))

    def set_damp(self, new_damp: int):
        if new_damp < 0 or new_damp > 10:
            raise ValueError("Dampers can only be set from 0 to 10")
        # Set to 0 is equivalent to turning off
        if new_damp == 0:
            self.turn_off()
        else:
            self.turn_on()
            damp_diff = new_damp - self.damp
            inc = damp_diff > 0
            for i in range(abs(damp_diff)):
                self.inc_dec_damp(inc)

    def _turn_on_off(self, on: bool):
        # there's a race here, should synchronise access to self.on
        if self.on != on:
            self._client.send_command(ToggleGroup(self.number))

    def turn_off(self):
        self._turn_on_off(False)

    def turn_on(self):
        self._turn_on_off(True)

    def get_status_strings(self):
        flags = [self.spill, self.turbo]
        flag_names = ['SPILL', 'TURBO']
        statuses = list(compress(flag_names, flags))
        if not statuses:
            statuses.append('NORMAL')
        return statuses

    def __str__(self):
        return f"""
        Group Name:\t{self.name}
        Group Number:\t{self.number}
        On:\t\t{self.on}
        Status:\t\t{self.get_status_strings()}
        Damper:\t\t{f'{self.damp*10}%'}
        """


import errno
import socket
import logging

from threading import Thread, Event
from queue import Empty, Queue
from time import sleep
from typing import Callable
from datetime import datetime

from airtouch2.AT2Aircon import AT2Aircon
from airtouch2.AT2Group import AT2Group

from airtouch2.protocol.constants import MessageLength
from airtouch2.protocol.messages import RequestState, ResponseMessage
from airtouch2.protocol.messages.CommandMessage import CommandMessage

from airtouch2.diff_bytes import print_diff_with_addresses

SocketClosedErrors = (errno.ECONNABORTED,  errno.ECONNRESET,
                      errno.ENOTCONN, errno.ESHUTDOWN, errno.ECONNREFUSED)

NetworkOrHostDownErrors = (errno.EHOSTUNREACH, errno.ECONNREFUSED,  errno.ETIMEDOUT, errno.ENETDOWN, errno.ENETUNREACH, errno.ENETRESET, errno.ECONNABORTED)

_LOGGER = logging.getLogger(__name__)

class AT2Client:

    def __init__(self, host: str, dump: bool=False):
        self._host_ip: str = host
        self._host_port: int = 8899
        self._sock: socket.socket = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM)
        self._sock.settimeout(None)
        self._stop_threads: bool = True
        self._cmd_queue: Queue[CommandMessage] = Queue()
        self.newest_response: ResponseMessage = None
        self._previous_response: bytes = None
        self._new_response: Event = Event()
        self._new_response_or_command: Event = Event()
        self.aircons: list[AT2Aircon] = []
        self.groups: list[AT2Group] = []
        self.system_name = "UNKNOWN"
        self._threads: list[Thread] = []
        self._active: bool = False
        self._callbacks: list[Callable] = []
        self._socket_broken = False
        self._data_updated = Event()
        self._dump = dump

    def __del__(self):
        if self._active:
            self.stop()

    def start(self) -> bool:
        if not self._connect():
            return False
        self._threads = [Thread(target=self._handle_incoming),
                         Thread(target=self._main_loop)]
        self._stop_threads = False
        for t in self._threads:
            t.start()
        self._active = True
        self.update_state()
        return True

    def _connect(self) -> bool:
        _LOGGER.debug(
            f'Connecting to {self._host_ip} on port {self._host_port}')
        try:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._sock.settimeout(None)
            self._sock.connect((self._host_ip, self._host_port))
        except OSError as e:
            _LOGGER.warning(f"Could not connect to host {self._host_ip}")
            if isinstance(e, socket.gaierror):
                # provided ip or port is rubbish/invalid
                pass
            elif e.errno not in NetworkOrHostDownErrors:
                raise e
            self._sock.close()
            return False
        _LOGGER.debug("Connected to airtouch2 server")
        return True

    def _shutdown_and_close_socket(self):
        try:
            self._sock.shutdown(socket.SHUT_RDWR)
        except OSError as e:
            if e.errno not in SocketClosedErrors:
                raise e
            _LOGGER.debug(
                f"Socket shutdown() call failed:\n  OSError: [Errno {e.errno}] {e.strerror}")
        self._sock.close()

    def _reset(self):
        # Originally in here I was also resetting self.aircons. In the context of this
        # client its fine, it will find them again when it gets a response message.
        # However the homeassistant entity is created with a reference to an aircon
        # which is then cleared so either it can't be cleared (hence why I removed it)
        # or there needs to be some callback that updates the homeassistant entity with
        # the new objet.
        self._new_response_or_command.clear()
        self._new_response.clear()
        self._data_updated.clear()
        self._socket_broken = False
        self.newest_response = None
        self.system_name = "UNKNOWN"
        # get rid of any commands on the queue
        while not self._cmd_queue.qsize() == 0:
            try:
                self._cmd_queue.get_nowait()
            except Empty:
                pass

    def stop(self):
        if self._active:
            self._stop_threads = True
            _LOGGER.debug("Closing socket...")
            self._shutdown_and_close_socket()
            self._data_updated.clear()
            self._new_response_or_command.set()
            _LOGGER.debug("Joining threads...")
            for t in self._threads:
                t.join()
            self._reset()
            self._threads.clear()
            self._active = False
            _LOGGER.debug("Shutdown successful")

        else:
            _LOGGER.debug("stop() called when client already inactive")

    def add_callback(self, func: Callable):
        self._callbacks.append(func)

        def remove_callback() -> None:
            if func in self._callbacks:
                self._callbacks.remove(func)

        return remove_callback

    def send_command(self, command: CommandMessage) -> None:
        self._cmd_queue.put(command)
        self._new_response_or_command.set()

    def update_state(self, block=True):
        self._data_updated.clear()
        self.send_command(RequestState())
        if block:
            self._data_updated.wait()

    def _reconnect(self) -> None:
        _LOGGER.debug("Reconnecting client...")
        self._shutdown_and_close_socket()
        self._reset()
        retries = 0
        while not self._connect():
            sleep(0.001 * (10**retries) if retries < 4 else 10)
            retries += 1
            if not retries % 50 or retries == 4:
                _LOGGER.debug(
                    "Server is not responding, will continue trying to reconnect every 10s")
        self.update_state(block=False)

    def _await_response(self) -> bytes:
        # taken (almost) straight from https://docs.python.org/3/howto/sockets.html
        chunks = []
        bytes_recd = 0
        while bytes_recd < MessageLength.RESPONSE:
            try:
                chunk = self._sock.recv(
                    MessageLength.RESPONSE - bytes_recd) if not self._stop_threads else b''
            except OSError as e:
                if e.errno not in SocketClosedErrors:
                    raise e
                chunk = b''
            if chunk == b'':
                # socket broken or closed
                break
            chunks.append(chunk)
            bytes_recd = bytes_recd + len(chunk)
        return b''.join(chunks)

    def _handle_incoming(self) -> None:
        while not self._stop_threads:
            while not self._socket_broken:
                resp = self._await_response()
                if len(resp) == 0:
                    self._socket_broken = True
                    break
                if len(resp) != MessageLength.RESPONSE:
                    _LOGGER.warning("Received invalid response, ignoring")
                    continue
                _LOGGER.debug("Received new valid response")
                self.newest_response = ResponseMessage(resp)
                self._new_response.set()
                self._new_response_or_command.set()
                if self._dump:
                    if self._previous_response:
                        print_diff_with_addresses(self._previous_response, resp)
                    with open('response_' + datetime.now().strftime("%m-%d-%Y_%H-%M-%S") + '.dump', 'wb') as f:
                        f.write(resp)
                self._previous_response = resp

            if not self._stop_threads and self._socket_broken:
                _LOGGER.debug("Socket connection was broken")
                self._reconnect()

    def _process_last_response(self):
        self._new_response.clear()
        self._new_response_or_command.clear()
        resp = self.newest_response
        self.newest_response = None
        # TODO: Check if dual unit system and handle 2 ACs
        # Aircons / units
        if not self.aircons:
            self.aircons.append(AT2Aircon(0, self, resp))
            _LOGGER.debug(self.aircons[0])
        else:
            for aircon in self.aircons:
                aircon.update(resp)
                _LOGGER.debug(aircon)
        # Groups
        if not self.groups or len(self.groups) != resp.num_groups:
            # TODO: callback to inform homeassistant
            self.groups.clear()
            for i in range(resp.num_groups):
                self.groups.append(AT2Group(self, i, resp))
        else:
            for group in self.groups:
                group.update(resp)
                _LOGGER.debug(group)
        self._data_updated.set()
        for func in self._callbacks:
            func()

    def _main_loop(self) -> None:
        """Main loop"""
        while not self._stop_threads:
            # wait for either a server message to be emitted or a command to be put on the queue
            _LOGGER.debug("Waiting for new data to arrive or command to send")
            self._new_response_or_command.wait()
            _LOGGER.debug("Main loop waiter Event triggered")
            # process a message if that's what triggered the event
            if self.newest_response:
                self._process_last_response()

            # process commands if that's what triggered the event
            while self._cmd_queue.qsize() > 0 and not self._socket_broken:
                _LOGGER.debug("There are commands in the queue")
                try:
                    cmd = self._cmd_queue.get(block=False)
                except Empty:
                    _LOGGER.debug("Tried to get command but queue was empty")
                    break
                self._new_response.clear()
                if not self._socket_broken:
                    _LOGGER.debug(f"Sending {cmd.__class__.__name__}")
                    self._sock.sendall(cmd.serialize())
                    self._new_response.wait(timeout=5)
                    if self.newest_response:
                        self._process_last_response()



from airtouch2.protocol.constants import CommandMessageConstants, CommandMessageType, GroupCommands
from airtouch2.protocol.messages.CommandMessage import CommandMessage

class GroupCommand(CommandMessage):
    """Base class from which all group control messages are derived"""
    def __init__(self, target_group_number: int):
        super().__init__()
        self.target_group = target_group_number

    def serialize(self, prefilled_msg: bytearray) -> bytearray:
        prefilled_msg[1] = CommandMessageType.GROUP_CONTROL
        prefilled_msg[3] = self.target_group

        return super().serialize(prefilled_msg)

class ToggleGroup(GroupCommand):
    """Command to toggle a group on or off"""

    def serialize(self) -> bytearray:
        serial_msg: bytearray = bytearray(self.length)
        serial_msg[4] = CommandMessageConstants.TOGGLE
        serial_msg[5] = GroupCommands.TOGGLE
        return super().serialize(serial_msg)

class ChangeDamper(GroupCommand):
    """Command to toggle a group on or off"""
    def __init__(self, target_group_number: int, inc: bool):
        super().__init__(target_group_number)
        self.inc = inc

    def serialize(self) -> bytearray:
        serial_msg: bytearray = bytearray(self.length)
        serial_msg[4] = GroupCommands.DAMP_INC if self.inc else GroupCommands.DAMP_DEC
        serial_msg[5] = GroupCommands.CHANGE_DAMP
        return super().serialize(serial_msg)


from airtouch2.protocol.enums import ACBrand

# I cannot find documentation of what these values actually mean
# The app prioritises this 'brand' (based on gateway id) over the other one in the message
# GATEWAY_BRAND2_LOOKUP = {
#     0x5  : 0x5,
#     0xFF : 0x5,
#     # 0x8: TETZUO (Daikin)
#     0x8  : 0x1,
#     # 0xD: me, gogreen, noirelec (fujitsu)
#     0xD  : 0x2,
#     0x22 : 0x2,
#     # 0xF: ruaandeysel (mitsubishi electric)
#     0xF  : 0x6,
#     0x10 : 0x4,
#     # 0x12 dmitry - (samsung)
#     0x12 : 0xE,
#     0x14 : 0xC,
#     0x15 : 0x7,
#     0x1F : 0xA,
#     0xE0 : 0xB,
#     0xE1 : 0xD
# }

# Lookup between the gateway ID and what brand I think it is
GATEWAYID_BRAND_LOOKUP = {
    0x8  : ACBrand.DAIKIN,
    0xD  : ACBrand.FUJITSU,
    0x22 : ACBrand.FUJITSU,
    0xF  : ACBrand.MITSUBISHI_ELECTRIC,
    0x12  : ACBrand.SAMSUNG
}